@@ -29,6 +29,7 @@
 #include "rsvg-structure.h"
 #include "rsvg-image.h"
 #include "rsvg-css.h"
+#include "string.h"
 
 #include <stdio.h>
 
@@ -87,10 +88,10 @@ _rsvg_node_init(RsvgNode *self)
 	self->children = g_ptr_array_new();
 	self->state = g_new(RsvgState, 1);
 	rsvg_state_init(self->state);
-	self->type = RSVG_NODE_PATH;
 	self->free = _rsvg_node_free;
 	self->draw = _rsvg_node_draw_nothing;
 	self->set_atts = _rsvg_node_dont_set_atts;
+	self->type = NULL;
 }
 
 void 
@@ -103,6 +104,8 @@ _rsvg_node_free (RsvgNode *self)
 		}
 	if (self->children != NULL)
 		g_ptr_array_free(self->children, TRUE);
+	if (self->type != NULL)
+		g_string_free(self->type, TRUE);
 	g_free (self);
 }
 
@@ -131,7 +134,6 @@ rsvg_new_group (void)
 	RsvgNodeGroup *group;
 	group = g_new (RsvgNodeGroup, 1);
 	_rsvg_node_init(&group->super);
-	group->super.type = RSVG_NODE_PATH;
 	group->super.draw = _rsvg_node_draw_children;
 	group->super.set_atts = rsvg_node_group_set_atts;
 	return &group->super;
@@ -159,75 +161,71 @@ rsvg_node_use_draw (RsvgNode * self, RsvgDrawingCtx *ctx,
 	RsvgNode * child;
 	RsvgState * state;
 	double affine[6];
+	double x, y, w, h;
+	x = _rsvg_css_normalize_length(&use->x, ctx, 'h');
+	y = _rsvg_css_normalize_length(&use->y, ctx, 'v');
+	w = _rsvg_css_normalize_length(&use->w, ctx, 'h');
+	h = _rsvg_css_normalize_length(&use->h, ctx, 'v');
 
 	rsvg_state_reinherit_top(ctx,  self->state, dominate);
 
-	rsvg_push_discrete_layer (ctx);
-
-
 	child = use->link;
 
 	/* If it can find nothing to draw... draw nothing */
 	if (!use->link)
 		return;
 	state = rsvg_state_current(ctx);	
-	if (child->type != RSVG_NODE_SYMBOL)
+	if (strcmp(child->type->str, "symbol"))
 		{
-			_rsvg_affine_translate(affine, use->x, use->y);
+			_rsvg_affine_translate(affine, x, y);
 			_rsvg_affine_multiply(state->affine, affine, state->affine);
 
+			rsvg_push_discrete_layer (ctx);
 			rsvg_state_push(ctx);	
 			rsvg_node_draw (child, ctx, 1);
 			rsvg_state_pop(ctx);
+			rsvg_pop_discrete_layer (ctx);
 		}
 	else
 		{
 			RsvgNodeSymbol *symbol = 
 				(RsvgNodeSymbol*)child;
-			double x, y, width, height;
-			x = use->x;
-			y = use->y;
-			width = use->w;
-			height = use->h;
 			
-			if (symbol->has_vbox){
+			if (symbol->vbox.active){
 				rsvg_preserve_aspect_ratio
 					(symbol->preserve_aspect_ratio, 
-					 symbol->width, symbol->height, 
-					 &width, &height, &x, &y);
-				
-				if (!symbol->overflow)
-					rsvg_add_clipping_rect (ctx, x, y, width, height);
+					 symbol->vbox.w, symbol->vbox.h, 
+					 &w, &h, &x, &y);
 
 				_rsvg_affine_translate(affine, x, y);
 				_rsvg_affine_multiply(state->affine, affine, state->affine);
-				_rsvg_affine_scale(affine, width / symbol->width,
-								   height / symbol->height);
+				_rsvg_affine_scale(affine, w / symbol->vbox.w,
+								   h / symbol->vbox.h);
 				_rsvg_affine_multiply(state->affine, affine, state->affine);
-				_rsvg_affine_translate(affine, -symbol->x, 
-									   -symbol->y);
+				_rsvg_affine_translate(affine, -symbol->vbox.x, 
+									   -symbol->vbox.y);
 				_rsvg_affine_multiply(state->affine, affine, state->affine);
 
+				_rsvg_push_view_box(ctx, symbol->vbox.w, symbol->vbox.h);
+				rsvg_push_discrete_layer (ctx);
+				if (!state->overflow || 
+					(!state->has_overflow && child->state->overflow))
+				rsvg_add_clipping_rect (ctx, symbol->vbox.x, symbol->vbox.y,
+										symbol->vbox.w, symbol->vbox.h);
 			} else {
-				_rsvg_affine_translate(affine, use->x, use->y);
+				_rsvg_affine_translate(affine, x, y);
 				_rsvg_affine_multiply(state->affine, affine, state->affine);
+				rsvg_push_discrete_layer (ctx);
 			}
 
 			rsvg_state_push(ctx);
 			_rsvg_node_draw_children(child, ctx, 1);
 			rsvg_state_pop(ctx);
-
+			rsvg_pop_discrete_layer (ctx);
+			if (symbol->vbox.active)
+				_rsvg_pop_view_box(ctx);
 		}
-	rsvg_pop_discrete_layer (ctx);
-}	
 
-static void 
-rsvg_node_use_free (RsvgNode *self)
-{
-	RsvgNodeUse *z = (RsvgNodeUse *)self;
-	rsvg_state_finalize (z->super.state);
-	g_free (z->super.state);
-	g_free (z);
 }
 
 static void
@@ -236,30 +234,38 @@ rsvg_node_svg_draw (RsvgNode * self, RsvgDrawingCtx *ctx,
 {
 	RsvgNodeSvg * sself;
 	RsvgState *state;
-	gdouble affine[6];
+	gdouble affine[6], affine_old[6], affine_new[6];
 	guint i;
+	double nx, ny, nw, nh;
 	sself = (RsvgNodeSvg *)self;
 
+	nx = _rsvg_css_normalize_length(&sself->x, ctx, 'h');
+	ny = _rsvg_css_normalize_length(&sself->y, ctx, 'v');
+	nw = _rsvg_css_normalize_length(&sself->w, ctx, 'h');
+	nh = _rsvg_css_normalize_length(&sself->h, ctx, 'v');
+
 	rsvg_state_reinherit_top(ctx, self->state, dominate);
 
-	rsvg_push_discrete_layer (ctx);
-
-	if (!sself->overflow && self->parent)
-		{
-			rsvg_add_clipping_rect(ctx, sself->x, sself->y, sself->w, sself->h);
-		}
 	state = rsvg_state_current (ctx);
 
-	if (sself->has_vbox && sself->hasw && sself->hash)
+	for (i = 0; i < 6; i++)
+		affine_old[i] = state->affine[i];
+
+	if (sself->vbox.active)
 		{
-			affine[0] = sself->w / sself->vbw;
+			double x = nx, y = ny, w = nw, h = nh;
+			rsvg_preserve_aspect_ratio(sself->preserve_aspect_ratio,
+									   sself->vbox.w, sself->vbox.h, 
+									   &w, &h, &x, &y);
+			affine[0] = w / sself->vbox.w;
 			affine[1] = 0;
 			affine[2] = 0;
-			affine[3] = sself->h / sself->vbh;
-			affine[4] = sself->x - sself->vbx * sself->w / sself->vbw;
-			affine[5] = sself->y - sself->vby * sself->h / sself->vbh;
+			affine[3] = h / sself->vbox.h;
+			affine[4] = x - sself->vbox.x * w / sself->vbox.w;
+			affine[5] = y - sself->vbox.y * h / sself->vbox.h;
 			_rsvg_affine_multiply(state->affine, affine, 
 								  state->affine);
+			_rsvg_push_view_box(ctx, sself->vbox.w, sself->vbox.h);
 		}
 	else
 		{
@@ -267,23 +273,38 @@ rsvg_node_svg_draw (RsvgNode * self, RsvgDrawingCtx *ctx,
 			affine[1] = 0;
 			affine[2] = 0;
 			affine[3] = 1;
-			affine[4] = sself->x;
-			affine[5] = sself->y;
+			affine[4] = nx;
+			affine[5] = ny;
 			_rsvg_affine_multiply(state->affine, affine, 
 								  state->affine);
+			_rsvg_push_view_box(ctx, nw, nh);
+		}
+	for (i = 0; i < 6; i++)
+		affine_new[i] = state->affine[i];
+
+	rsvg_push_discrete_layer (ctx);
+
+	/* Bounding box addition must be AFTER the discrete layer push, 
+	   which must be AFTER the transformation happens. */
+	if (!state->overflow && self->parent)
+		{
+			for (i = 0; i < 6; i++)
+				state->affine[i] = affine_old[i];
+			rsvg_add_clipping_rect(ctx, nx, ny, nw, nh);
+			for (i = 0; i < 6; i++)
+				state->affine[i] = affine_new[i];
 		}
 
 	for (i = 0; i < self->children->len; i++)
 		{
 			rsvg_state_push(ctx);
-
 			rsvg_node_draw (g_ptr_array_index(self->children, i), 
 									 ctx, 0);
-	
 			rsvg_state_pop(ctx);
 		}			
 
 	rsvg_pop_discrete_layer (ctx);
+	_rsvg_pop_view_box(ctx);
 }
 
 static void
@@ -297,43 +318,23 @@ rsvg_node_svg_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag *atts)
 	if (rsvg_property_bag_size (atts))
 		{
 			if ((value = rsvg_property_bag_lookup (atts, "viewBox")))
-				{
-					svg->has_vbox = rsvg_css_parse_vbox (value, 
-														 &svg->vbx, 
-														 &svg->vby,
-														 &svg->vbw, 
-														 &svg->vbh);
-					if (svg->has_vbox)
-						{
-							ctx->width = svg->vbw;
-							ctx->height = svg->vbh;
-						}
-				}
+				svg->vbox = rsvg_css_parse_vbox (value);
+			
+			if ((value = rsvg_property_bag_lookup (atts, "preserveAspectRatio")))
+				svg->preserve_aspect_ratio = rsvg_css_parse_aspect_ratio (value);			
 			if ((value = rsvg_property_bag_lookup (atts, "width")))
-				{
-					svg->w = rsvg_css_parse_normalized_length (value, ctx->dpi_x, ctx->width, 1);
-					svg->hasw = (svg->w > 0);
-					if (!svg->has_vbox)
-						ctx->width = svg->w; 
-				}
+				svg->w = _rsvg_css_parse_length (value);
 			if ((value = rsvg_property_bag_lookup (atts, "height")))
-				{
-					svg->h = rsvg_css_parse_normalized_length (value, ctx->dpi_y, ctx->height, 1);
-					svg->hash = (svg->h > 0);
-					if (!svg->has_vbox)
-						ctx->height = svg->h;
-				}
+				svg->h = _rsvg_css_parse_length (value);
 			if ((value = rsvg_property_bag_lookup (atts, "x")))
-				svg->x = rsvg_css_parse_normalized_length (value, ctx->dpi_x, ctx->width, 1);
+				svg->x = _rsvg_css_parse_length (value);
 			if ((value = rsvg_property_bag_lookup (atts, "y")))
-				svg->y = rsvg_css_parse_normalized_length (value, ctx->dpi_y, ctx->height, 1);
+				svg->y = _rsvg_css_parse_length (value);
 			if ((value = rsvg_property_bag_lookup (atts, "id")))
 				{
 					id = value;
 					rsvg_defs_register_name (ctx->defs, value, &svg->super);
 				}
-			if ((value = rsvg_property_bag_lookup (atts, "overflow")))
-				svg->overflow = rsvg_css_parse_overflow(value);
 		}
 }
 
@@ -343,15 +344,14 @@ rsvg_new_svg (void)
 	RsvgNodeSvg * svg;
 	svg = g_new (RsvgNodeSvg, 1);
 	_rsvg_node_init(&svg->super);
-	svg->has_vbox = FALSE;
+	svg->vbox.active = FALSE;
 	svg->preserve_aspect_ratio = RSVG_ASPECT_RATIO_XMID_YMID;
-	svg->x = 0; svg->y = 0; svg->w = -1; svg->h = -1;
-	svg->hasw = svg->hash = FALSE;
-	svg->vbx = 0; svg->vby = 0; svg->vbw = 0; svg->vbh = 0;
-	svg->super.type = RSVG_NODE_PATH;
+	svg->x = _rsvg_css_parse_length ("0"); 
+	svg->y = _rsvg_css_parse_length ("0"); 
+	svg->w = _rsvg_css_parse_length ("100%");
+	svg->h = _rsvg_css_parse_length ("100%");
 	svg->super.draw = rsvg_node_svg_draw;
 	svg->super.set_atts = rsvg_node_svg_set_atts;
-	svg->overflow = FALSE;
 	return &svg->super;
 }
 
@@ -359,22 +359,19 @@ static void
 rsvg_node_use_set_atts (RsvgNode * self, RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
 	const char *value = NULL, *klazz = NULL, *id = NULL;
-	double font_size;	
 	RsvgNodeUse * use;
 
-	font_size = rsvg_state_current_font_size(ctx);
-
 	use = (RsvgNodeUse *)self;
 	if (rsvg_property_bag_size(atts))
 		{
 			if ((value = rsvg_property_bag_lookup (atts, "x")))
-				use->x = rsvg_css_parse_normalized_length (value, ctx->dpi_x, (gdouble)ctx->width, font_size);
+				use->x = _rsvg_css_parse_length(value);
 			if ((value = rsvg_property_bag_lookup (atts, "y")))
-				use->y = rsvg_css_parse_normalized_length (value, ctx->dpi_y, (gdouble)ctx->height, font_size);
+				use->y = _rsvg_css_parse_length(value);
 			if ((value = rsvg_property_bag_lookup (atts, "width")))
-				use->w = rsvg_css_parse_normalized_length (value, ctx->dpi_x, (gdouble)ctx->height, font_size);
+				use->w = _rsvg_css_parse_length(value);
 			if ((value = rsvg_property_bag_lookup (atts, "height")))
-				use->h = rsvg_css_parse_normalized_length (value, ctx->dpi_y, (gdouble)ctx->height, font_size);
+				use->h = _rsvg_css_parse_length(value);
 			if ((value = rsvg_property_bag_lookup (atts, "class")))
 				klazz = value;
 			if ((value = rsvg_property_bag_lookup (atts, "id")))
@@ -395,14 +392,12 @@ rsvg_new_use ()
 	RsvgNodeUse * use;
 	use = g_new (RsvgNodeUse, 1);
 	_rsvg_node_init(&use->super);
-	use->super.type = RSVG_NODE_PATH;
-	use->super.free = rsvg_node_use_free;
 	use->super.draw = rsvg_node_use_draw;
 	use->super.set_atts = rsvg_node_use_set_atts;
-	use->x = 0;
-	use->y = 0;
-	use->w = 0;
-	use->h = 0;
+	use->x = _rsvg_css_parse_length("0");
+	use->y = _rsvg_css_parse_length("0");
+	use->w = _rsvg_css_parse_length("0");
+	use->h = _rsvg_css_parse_length("0");
 	use->link = NULL;
 	return (RsvgNode *)use;
 }
@@ -424,22 +419,10 @@ rsvg_node_symbol_set_atts(RsvgNode *self, RsvgHandle *ctx, RsvgPropertyBag *atts
 					rsvg_defs_register_name (ctx->defs, value, &symbol->super);
 				}
 			if ((value = rsvg_property_bag_lookup (atts, "viewBox")))
-				{
-					symbol->has_vbox = rsvg_css_parse_vbox (value, 
-															&symbol->x, 
-															&symbol->y,
-															&symbol->width, 
-															&symbol->height);
-					if (symbol->has_vbox)
-						{
-							ctx->width = symbol->width;
-							ctx->height = symbol->height;
-						}
-				}
+				symbol->vbox = rsvg_css_parse_vbox (value);
 			if ((value = rsvg_property_bag_lookup (atts, "preserveAspectRatio")))
-				symbol->preserve_aspect_ratio = rsvg_css_parse_aspect_ratio (value);			
-			if ((value = rsvg_property_bag_lookup (atts, "overflow")))
-				symbol->overflow = rsvg_css_parse_overflow(value);
+				symbol->preserve_aspect_ratio = 
+					rsvg_css_parse_aspect_ratio (value);			
 		}
 
 	rsvg_parse_style_attrs (ctx, self->state, "symbol", klazz, id, atts);
@@ -452,10 +435,8 @@ rsvg_new_symbol(void)
 	RsvgNodeSymbol * symbol;
 	symbol = g_new (RsvgNodeSymbol, 1);
 	_rsvg_node_init(&symbol->super);
-	symbol->has_vbox = 0;
-	symbol->overflow = 0;
+	symbol->vbox.active = FALSE;
 	symbol->preserve_aspect_ratio = RSVG_ASPECT_RATIO_XMID_YMID;
-	symbol->super.type = RSVG_NODE_SYMBOL;
 	symbol->super.draw = _rsvg_node_draw_nothing;
 	symbol->super.set_atts = rsvg_node_symbol_set_atts;
 	return &symbol->super;
@@ -467,7 +448,6 @@ rsvg_new_defs ()
 	RsvgNodeGroup *group;
 	group = g_new (RsvgNodeGroup, 1);
 	_rsvg_node_init(&group->super);
-	group->super.type = RSVG_NODE_PATH;
 	group->super.draw = _rsvg_node_draw_nothing;
 	group->super.set_atts = rsvg_node_group_set_atts;
 	return &group->super;
@@ -506,7 +486,6 @@ rsvg_new_switch (void)
 	RsvgNodeGroup *group;
 	group = g_new (RsvgNodeGroup, 1);
 	_rsvg_node_init(&group->super);
-	group->super.type = RSVG_NODE_PATH;
 	group->super.draw = _rsvg_node_switch_draw;
 	group->super.set_atts = rsvg_node_group_set_atts;
 	return &group->super;
